import argparse

import prettytensor as pt

import samyro.learn
import samyro.model
import samyro.read
import samyro.write

import samyro.cli.learn
import samyro.cli.write
import samyro.cli.sample

from samyro.cli import positive_int


def get_model(args):
    """Summarize model according to user specified params."""
    return samyro.model.Model(name=args.name,
                              embedding=args.embedding_size,
                              lstms=args.lstm_dims)


def get_runner(args):
    return pt.train.Runner(save_path=args.checkpoint_pattern)


def set_shared_args():
    """All subcommands use these arguments."""

    # TODO: add a verbosity level.
    shared_parent = argparse.ArgumentParser(add_help=False)

    # Model config arguments
    model_group = shared_parent.add_argument_group('model configuration')
    model_group.add_argument(
        '--embedding_size',
        metavar="N_DIMS",
        help="size of input character embedding",
        default=16,
        type=int)

    model_group.add_argument(
        "--lstm_dims", metavar="STACK_OF_LSTM_DIMS",
        help="comma-separated list of LSTM sizes",
        type=lambda x: tuple(int(y.strip()) for y in x.split(",")),
        default=(128, 256))

    model_group.add_argument(
        "--name", metavar="namespace prefix",
        help="namespace prefix for this model; stored internally",
        default='shakespeare', type=str)

    checkpoint_group = shared_parent.add_argument_group(
        'parameter configuration')

    checkpoint_group.add_argument(
        '--checkpoint_pattern',
        metavar="FILE_PREFIX",
        help="where to find checkpoints",
        type=str,
        default='/home/jeremy/notebooks/shakespeare/shakespeare')

    return shared_parent


def set_sampler_shared_args(parser, default_batch_size=8):
    reader_group = parser.add_argument_group(
        'sampler read configuration')

    reader_group.add_argument(
        '--batch_size', type=positive_int, default=default_batch_size,
        help="how many segments should be read per backprop batch")

    reader_group.add_argument(
        '--steps_per_sample',
        type=positive_int, default=100,
        help="how many timesteps per sample")

    reader_group.add_argument(
        '--sampler',
        choices=['patches', 'lines', 'paragraphs'],
        default='patches',
        help="select the class of sampler to use on the file")


def get_sampler(args):
    if args.sampler == 'patches':
        sampler_class = samyro.read.CharacterPatchSampler
    elif args.sampler == 'lines':
        sampler_class = samyro.read.LineSampler
    elif args.sampler == 'paragraphs':
        sampler_class = samyro.read.ParagraphSampler
    else:
        assert False, "unrecognized --sampler %s" % args.sampler

    return sampler_class(
        filename=args.file.name,
        sample_length=args.steps_per_sample,
        batch_size=args.batch_size)


def main():
    """Entry point for main samyro script with subcommands."""

    shared_parent = set_shared_args()

    # Main argument parsers.
    parser = argparse.ArgumentParser(prog='samyro', fromfile_prefix_chars='@')

    subparsers = parser.add_subparsers(help="samyro subcommands",
                                       dest="subcommand_name")

    learner_parser = subparsers.add_parser("learn", help="reading a text",
                                           parents=[shared_parent])

    samyro.cli.learn.set_learner_args(learner_parser)

    writer_parser = subparsers.add_parser("write", help="generate new text",
                                          parents=[shared_parent])
    samyro.cli.write.set_writer_args(writer_parser)

    sample_parser = subparsers.add_parser("sample",
                                          help="debug sampling from files")
    samyro.cli.sample.set_sampler_args(sample_parser)

    # Actual work done by the execute function declared inside each subcommand

    # TODO: This would be a great place to put a shared call to argcomplete.
    args = parser.parse_args()

    args.execute(args)



"""Utility functions and definitions for converting to/from integers."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


import random

import tensorflow as tf

UNK = 0
BOS = 127
EOS = 1

UNK_CHAR = u'\ufeff'
BOS_CHAR = "\x7f"
EOS_CHAR = "\x01"

CHARS = 128


def placeholder(batch_size, timesteps):
    return tf.placeholder(tf.int32, [batch_size, timesteps])


def convert_to_int(char, ordmax=128):
    i = ord(char)
    if i >= ordmax:
        return UNK
    return i


def random_seed_char():
    return chr(ord('A') + random.randint(0, 25))


def convert_to_symbol(integer, ordmax=128):
    if integer >= ordmax:
        return UNK_CHAR
    return chr(integer)



"""Defines models for reading that save out to checkpoints."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from samyro import integerize

import prettytensor as pt
import tensorflow as tf

from prettytensor.tutorial import data_utils


class Model(object):
    """A model shape.

    TODO(jkahn): convert to collections.namedtuple?
    """
    def __init__(self, name, embedding=16, lstms=(128, 256)):
        self.name = name
        self.embedding = embedding
        self.lstms = lstms

    def create(self, input_placeholder, phase):
        """Creates a 2 layer LSTM model with dropout.

        Args:
          input_placeholder: placeholder of timesteps x sequences

          phase: Phase controls whether or not dropout is active.  In
            training mode we want to perform dropout, but in test we
            want to disable it.

        Returns: The logits layer.

        """
        timesteps = input_placeholder.get_shape()[1].value
        text_in = data_utils.reshape_data(input_placeholder)
        with pt.defaults_scope(activation_fn=tf.nn.relu, l2loss=0.00001):
            # The embedding lookup must be placed on a cpu.
            with tf.device('/cpu:0'):
                embedded = text_in.embedding_lookup(integerize.CHARS,
                                                    [self.embedding])
        # Because the sequence LSTM expects each timestep to be its
        # own Tensor, we need to cleave the sequence.  Below we can
        # build a stacked 2 layer LSTM by just chaining them together.
        # You can stack as many layers as you want.
        lstm = embedded.cleave_sequence(timesteps)
        assert len(self.lstms)
        for lstm_size in self.lstms:
            lstm = lstm.sequence_lstm(lstm_size)

        # The classifier is much more efficient if it runs across the entire
        # dataset at once, so we want to squash (i.e. uncleave).
        # Note: if phase is test, dropout is a noop.
        return (lstm.squash_sequence()
                .dropout(keep_prob=0.8, phase=phase)
                .fully_connected(integerize.CHARS, activation_fn=None))

    def train_op_loss(self, input_placeholder, labels, reuse=None):
        # Training and eval graph
        with tf.variable_scope(self.name, reuse=reuse):
            # Core train graph
            result = self.create(input_placeholder,
                                 pt.Phase.train).softmax(labels)

            train_op = pt.apply_optimizer(tf.train.AdagradOptimizer(0.5),
                                          losses=[result.loss])
            return train_op, result.loss

    def eval_accuracy(self, input_placeholder, labels, reuse=True):
        with tf.variable_scope(self.name, reuse=reuse):
            # Eval graph
            eval_result = self.create(input_placeholder,
                                      pt.Phase.test).softmax(labels)

        # Accuracy creates variables, so do it outside of scope
        return eval_result.softmax.evaluate_classifier(labels,
                                                       phase=pt.Phase.test)

    def inference_io(self, reuse):
        with tf.variable_scope(self.name, reuse=reuse), pt.defaults_scope(
                summary_collections=['INFERENCE_SUMMARIES']):
            inf_input = tf.placeholder(tf.int32, [])
            inf_logits = self.create(pt.wrap(inf_input).reshape([1, 1]),
                                     pt.Phase.infer)
            return inf_input, inf_logits



"""Test reading batches of character streams."""
from samyro import read

from cStringIO import StringIO

import numpy


def test_patch_batches():
    rr = read.CharacterPatchSampler(filehandle=StringIO(u'foobar\n'),
                                    sample_length=3, batch_size=2)
    batches = rr.batches(shuffle=False)

    first = batches.next()
    assert isinstance(first, read.Sample)
    assert isinstance(first.input, numpy.ndarray)

    batches = rr.batches(shuffle=False, to_numpy=False)

    first = list(batches.next())
    assert len(first) == 2
    assert first[0] == read.Sample(input="\x7ffo", output="foo")


def test_patch_strings():
    rr = read.CharacterPatchSampler(filehandle=StringIO(u'foobar\n'),
                                    sample_length=3, batch_size=2)
    strings = rr.string_samples()

    first = strings.next()
    assert isinstance(first, read.Sample)
    assert read.Sample(input="\x7ffo", output="foo") == first

    second = strings.next()
    assert isinstance(second, read.Sample)
    assert read.Sample(input="oba", output="bar") == second

    third = strings.next()
    assert isinstance(third, read.Sample)
    assert read.Sample(input="r\n\x01", output="\n\x01\x01") == third

    # assert raises EndOfInput thereafter?


def test_lines_strings():
    rr = read.LineSampler(filehandle=StringIO(u'foobar\nbaz\nwikiwiki\n'),
                          sample_length=6, batch_size=2)
    strings = rr.string_samples()

    first = strings.next()
    assert isinstance(first, read.Sample)
    assert read.Sample(input="\x7ffooba", output="foobar") == first

    second = strings.next()
    assert isinstance(second, read.Sample)
    assert read.Sample(input="\x7fbaz\n\x01", output="baz\n\x01\x01") == second

    third = strings.next()
    assert isinstance(third, read.Sample)
    assert read.Sample(input="\x7fwikiw", output="wikiwi") == third

    # assert raises EndOfInput thereafter?


def test_paras_strings():
    rr = read.ParagraphSampler(
        filehandle=StringIO(u'foobar\nbaz\n\n\nwikiwiki\nyap\n\nwooo'),
        sample_length=15, batch_size=2)
    strings = rr.string_samples()

    first = strings.next()
    assert isinstance(first, read.Sample)
    assert read.Sample(input="\x7ffoobar\nbaz\n\x01\x01\x01",
                       output="foobar\nbaz\n\x01\x01\x01\x01") == first

    second = strings.next()
    assert isinstance(second, read.Sample)
    assert read.Sample(input="\x7fwikiwiki\nyap\n\x01",
                       output="wikiwiki\nyap\n\x01\x01") == second

    third = strings.next()
    assert isinstance(third, read.Sample)
    assert read.Sample(input="\x7fwooo" + "\x01" * 10,
                       output="wooo" + "\x01" * 11) == third

    # assert raises EndOfInput thereafter?



import argparse
import tensorflow as tf

from samyro.cli import positive_int, positive_float
import samyro.cli.shared
import samyro.integerize


def execute(args):
    """Actual execution of the model in the checkpoint_pattern."""
    model = samyro.cli.shared.get_model(args)
    inference_i, inference_o = model.inference_io(reuse=None)
    with tf.Session() as sess:
        runner = samyro.cli.shared.get_runner(args)
        runner.load_from_checkpoint(sess)
        text = samyro.write.write(inference_i, inference_o,
                                  seed=args.seed,
                                  max_length=args.max_length,
                                  temperature=args.temperature)
        print(text)


def set_writer_args(writer_parser):
    """The write subcommand uses the shared arguments but also these."""

    writer_parser.add_argument('--seed', type=str,
                               default=samyro.integerize.BOS_CHAR,
                               help="seed string to bootstrap hidden state")

    writer_parser.add_argument('--max_length', type=positive_int, default=1024,
                               help="max length in characters")

    writer_parser.add_argument('--temperature', type=positive_float,
                               default=0.5,
                               help="temperature (positive float)")

    writer_parser.set_defaults(execute=execute)


def main():
    """Entry point for main samyro-write script."""

    shared_parent = samyro.cli.shared.set_shared_args()

    parser = argparse.ArgumentParser(prog='samyro-write',
                                     fromfile_prefix_chars='@',
                                     parents=[shared_parent])

    set_writer_args(parser)

    args = parser.parse_args()

    execute(args)



"""Samyro -- an automatic Dada-ist poet.

Designed to read text and write more texts "just like them".

"Samyro" was the first pseudonym of Samuel (Samy) Rosenstock,
better known by his later pseudonym "Tristan Tzara".
"""

# Always prefer setuptools over distutils
from setuptools import setup, Command, find_packages
# To use a consistent encoding
from codecs import open
from os import path
from subprocess import call


here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


class RunTests(Command):
    """Run all tests."""
    description = 'run tests'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        """Run all tests!"""
        errno = call(['py.test', 'tests'])
        raise SystemExit(errno)

setup(
    name='samyro',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.0.2',

    description='An automatic text generator based on RNNs.',
    long_description=long_description,

    # The project's main homepage.
    url='http://trochee.net/samyro',

    # Author details
    author='Jeremy G. Kahn',
    author_email='jeremy@trochee.net',

    # Choose your license
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 2 - Pre-Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Topic :: Communications',
        'Topic :: Education',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Text Processing :: Linguistic',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],

    # What does your project relate to?
    keywords='RNNs poetry generation',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['six',
                      'tensorflow>=0.7.1',
                      'prettytensor>=0.5.3',
    ],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'dev': ['check-manifest', 'flake8', 'jupyter', 'twine'],
        'test': ['coverage', 'check-manifest', 'flake8', 'docutils', 'pytest'],
    },

    # # If there are data files included in your packages that need to be
    # # installed, specify them here.  If using Python 2.6 or less, then these
    # # have to be included in MANIFEST.in as well.
    # package_data={
    #     'sample': ['package_data.dat'],
    # },

    # # Although 'package_data' is the preferred approach, in some case you may
    # # need to place data files outside of your packages. See:
    # # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[('my_data', ['data/data_file'])],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'samyro=samyro.cli.shared:main',
            'samyro-learn=samyro.cli.learn:main',
            'samyro-sample=samyro.cli.sample:main',
            'samyro-write=samyro.cli.write:main',
        ],
    },
    cmdclass={'test': RunTests},
)
